#! /usr/bin/python
# -*- coding: utf-8 -*-


import sys, re, time, getopt, nltk, math, os, string, Levenshtein, codecs, profile
from operator import itemgetter, attrgetter


#####################################################
# GLOBAL VARIABLES

_SPECTagged = ""   # Specialized tagged and lemmatized (TreeTagger) corpus
_split = ""           # List od specialized German compounds with their splitting 
_lang = ""               # Language
_rules = ""              # File with rules

_out = "tagged.new" # Output file

#####################################################
# Getting programme options
def getOpts(argv):

   def usage():
      print "Usage: python ", sys.argv[0], "-t tagged-specialized-corpus -l language -s splitted compounds -r file-with-rules"

   opts, args = getopt.getopt(argv, "hs:t:l:s:r:", ["--help", "external-vocabulary=", "spec-tagged=", "language=", "splitted=", "rules="])
   #print opts
   for opt, arg in opts:
        #print opt, arg
        if opt in ("-h", "--help"):
            usage()                     
            sys.exit()                  
        elif opt in ('-e', "--external-vocabulary"):
            global _extVocab
            _extVocab = arg      
        elif opt in ('-t', "--spec-tagged"):
            global _SPECTagged          
            _SPECTagged = arg   
        elif opt in ("-l", "--language"):
           global _lang
           _lang = arg
           print "Language: ", _lang
        elif opt in ("-s", "--splitted"):
           global _split
           _split = arg
        elif opt in ("-r", "--rules"):
           global _rules
           _rules = arg

##################################
class WordPosLemmas: 

  # Konstruktor:
  # wordform = word form
  # pos = POS
  # lemmas = dictionary of lemmas with pos and frequency
  def __init__(self, value, pos):            
    self.wordform = value
    self.pos = pos  
    self.lemmas = {}
    self.new_lemmas = {}
    self.head = value
    self.split_list = []
    self.splits = {}
    self.compLemmaFound = False

  def addLemma(self, lemma, pos):
     if pos not in self.lemmas:
        self.lemmas[pos] = {}
        self.lemmas[pos][lemma] = 1 # {lemmas: {'N': {'Auto': 1}, ...}, ...}
     else:
        if lemma not in self.lemmas[pos]:
           self.lemmas[pos][lemma] = 1 # {lemmas: {'N': {'Auto': 1}, ...}, ...}
        else:
           self.lemmas[pos][lemma] += 1
     #print self.lemmas

  def addLemmasFrom(self, new):
     for pos in self.lemmas:
        if pos in new.lemmas:
           for lemma in new.lemmas[pos]:
              try:
                 self.lemmas[pos][lemma] += 1
              except KeyError:
                 self.lemmas[pos][lemma] = 1

  def addPosLemma(self, pos, lemma):
     self.new_lemmas[pos] = lemma

  def addSplits(self, split_list, wf_lemma_dict):

     #print "splits = ", split_list
     new_head = None

     def checkSplits(split_list):
        ok = True
        for s in split_list:
           if s in ['', '-']:
              ok = False
              break
        return ok

     if len(split_list) > 1 and checkSplits(split_list):
        for s in xrange(len(split_list)):


           if s == len(split_list)-1: # Head noun
              self.head = split_list[s]

              try:
                 wf_lemma_dict[split_list[s]].addLemma(split_list[s], "N----") # Wordform already exists
                 self.splits[split_list[s]] = wf_lemma_dict[split_list[s]]
                 #print "Schon vorhanden", self.splits[split_list[s]].lemmas
                 self.split_list.append(split_list[s])
              except:
                 new_wf =  WordPosLemmas(split_list[s], "N----") # New wordform
                 new_wf.addLemma(split_list[s], "N----")
                 self.splits[split_list[s]] = new_wf
                 new_head = self.splits[split_list[s]]
                 self.split_list.append(split_list[s])

           #self.splits[s].printObject()

     return (wf_lemma_dict, new_head)
             

  def isCompound(self):

     if self.splits != {}:
        return True
     else:
        return False


  def createCompoundLemma(self):

     res = None
     #print "-----"
     #self.printObject()

     pos_n = ['Ncfsn', 'Ncfsg', 'Ncfsd', 'Ncfsa', 'Ncmsn', 'Ncmsd', 'Ncmsa', 'Ncnsn', 'Ncnsd', 'Ncnsa']
     
     pos_found = False
     if 'N' in self.new_lemmas:
        lemma_norm = normalizeWord(self.new_lemmas['N'], "de")
        wf_norm = normalizeWord(self.wordform, "de")
        #print "lemma_norm = ", lemma_norm 
        #print "wf_norm = ", wf_norm 
        if lemma_norm in wf_norm:
           if len(lemma_norm) < len(wf_norm):
              pos_found = True
           else:
              for p in self.lemmas:
                 if p in pos_n:
                    pos_found = True
                    break

     try:

     # Only if the new lemma is not derived from 
     # the corresponding pos tag
        if not pos_found:

          # print "Neuberechnung"

           head_lemma = self.splits[self.head].new_lemmas['N']
           if head_lemma.startswith("-"): 
              head_lemma = head_lemma[1:]
              #print "neues head-lemma ", head_lemma, " fuer ", self.wordform
     
           head_start_idx = self.wordform.lower().find(self.head.lower())
           res = self.wordform[0:head_start_idx]

           if head_start_idx != 0 and head_start_idx != -1:

              if res[-1] == "-":
                 res += head_lemma[0].upper() + head_lemma[1:].lower()
              else:
                 res += head_lemma[0].lower() + head_lemma[1:].lower()

           #print "Neues CP-Lemma = ", res

        else:

           res = self.new_lemmas['N']

     except Exception as inst:
        res = self.new_lemmas['N']
   
     return res

  def printObject(self):

     def printComp(self):
         for s in self.splits:
            print "\tWord: ", self.splits[s].wordform
            print "\tAll lemmas: ", self.splits[s].lemmas
            print "\tNew lemma(s): ", self.splits[s].new_lemmas 

     print "Word: ", self.wordform
     print "All lemmas: ", self.lemmas
     print "New lemma(s): ", self.new_lemmas 
     if self.isCompound():
        print "Compound head: ", self.head
        print "Compound parts: ", self.split_list
        printComp(self)      
     print "\n"

# Reading in the tagged input. 
# Input: word form \t POS \t lemma
# Output: dictionary of word forms with POS, lemmas and their frequencies 
# E.g. {turbines: {'N': {turbine: 25, turbine:3}, ...}, ...}
def getLemmaWordListWithPOSFromTagged(tagged, wf_lemma_dict):

   res = {}
   wl_file = open(tagged, "r")

   wl_line = wl_file.readline()
   line_nr = 1

   print "Reading lexicon from the tagged file..."
   while wl_line:

      #print wl_line, " ", wl_line.split("\t")
      #print line_nr, "\r", 

      try:

         if not wl_line.startswith("<") and not wl_line.endswith(">"):
         
            wf = wl_line.split("\t")[0]
            if wf != '':
              #print "WF = ", wf

               #pos = wl_line.split("\t")[1].strip().split(".")[0] # POS from RFTagger
               #pos = wl_line.split("\t")[1].strip()[0] # POS from MulText
               pos = wl_line.split("\t")[1].strip() # POS + morph. from MulText

               if pos[0] in ['N', 'V', 'A']:

                  lemma = wl_line.split("\t")[2].strip()
                  #print type(lemma)
                  if lemma == "<unknown>": # Replace unknown lemmas by word form
                     lemma = wf
                    #print "Lemma = ", lemma

                  lemmas = lemma.split("|") # Ambiguous lemmas are split
                  #print "Lemmas = ", lemmas
               
                  try:
                  #print "Adding lemmas to existing instance"
                     for l in lemmas:
                        wf_lemma_dict[wf].addLemma(l, pos)
         
                  except KeyError:
                     new_wf =  WordPosLemmas(wf, pos)
                     for l in lemmas:
                        new_wf.addLemma(l, pos)
                     wf_lemma_dict[wf] = new_wf
      
         line_nr += 1
         wl_line = wl_file.readline()

      except Exception as inst:
         print "Error at ", wl_line, " : ", inst, ". Skipping the line"
         wl_line = wl_file.readline()

   print "Dictionary size: ", len(wf_lemma_dict) #, wf_lemma_dict
   return wf_lemma_dict

# Reads a list of split compounds
# Input: Noun \t Split1|Split2
def getWordsFromSplit(split, wf_lemma_dict_old):

   wf_lemma_dict = wf_lemma_dict_old.copy()

   split_file = open(split, "r")
   split_line = split_file.readline()

   while split_line:

      #print split_line.strip().split("\t")
      (wordform, split) = split_line.strip().split("\t")

      try:
         (wf_lemma_dict, new_head) = wf_lemma_dict[wordform.strip()].addSplits(split.split("|"), wf_lemma_dict)

         if new_head != None:
            #print "neues wort ", new_head.wordform, " ", new_head
            wf_lemma_dict[new_head.wordform] = new_head
            #wf_lemma_dict[new_head.wordform].printObject()
   
      except Exception as inst:
         #print "Ueberfluessiges Wort ", wordform
         pass

      split_line = split_file.readline()
    
   split_file.close()
   del wf_lemma_dict_old
   print "Size of enlarged dictionary ", len(wf_lemma_dict)
   return wf_lemma_dict


def normalizeWord(word, lang):

   res = ""
   #print "Word: ", word # Eingelesenes Wort
   word.decode('utf-8')
   word2 = unicode(word, 'utf-8') # Konvertierung nach unicode
   word3 = word2.encode('utf-8') # Konvertierung nach UTF-8 zum Ausgeben in der Konsole
      #print "T0 ", word3, " ", type(word2)

   # Language-independent normalization:
   # Removing hyphen, lowecasing the word
   def punct(char):

      res = ""
      spec = [u'à', u'á', u'è', u'é', u'í', u'ò', u'ó', u'ñ', u'ä', u'ü', u'ö', u'ß']
     
      try:
         if char in spec or char == "-":
            return char
         else:
            if char.isalpha():
               res =  char.lower()
            else:
               return ""

      except Exception as inst:
         print "Zeichen nicht ok"
         
      #print "R", type(res)
      return res

   def replaceSpecChars(char):

      #print char, " ", type(char)
            
      reps = {u'ä':'a', u'ö':'o', u'ü':'u', u'Ä':'A', u'Ö':'O', u'Ü':'U', u'ß':'ss', u'à': 'a', u'á': 'a', u'è':'e', u'é':'e', u'í':'i', u'ò': u'o', u'ó':'o', u'ñ': 'n'}
      #reps = {u'à': 'a', u'á': 'a', u'è':'e', u'é':'e', u'í':'i', u'ò': u'o', u'ó':'o', u'ñ': 'n'}

      #print reps
      if char in reps:
         char = reps[char]

      #print "Res ", char
      return char


   c = [None]*len(word2) 
   for i in xrange(len(word2)): 
      c[i] = replaceSpecChars(punct(word2[i].lower()))
   res = "".join(c)
   if res.startswith("-"):
      res = res[1:]
   elif res.endswith("-"):
      res = res[0:-1]

   #print res
   res = res.encode('utf-8')
   #print "ferig ", res
   return res

def normalizeWordForRules(word, lang):

      res = ""
      #print word, " ", type(word) # Eingelesenes Wort
      word.decode('utf-8')
      word2 = unicode(word, 'utf-8') # Konvertierung nach unicode
      word3 = word2.encode('utf-8') # Konvertierung nach UTF-8 zum Ausgeben in der Konsole
      #print "T0 ", word3, " ", type(word2)

      # Language-independent normalization:
      # Removing hyphen, lowecasing the word
      def punct(char):

          res = ""
          spec = [u'à', u'á', u'è', u'é', u'í', u'ò', u'ó', u'ñ']
     
          try:
             if char in spec or char == "-":
                return char
             else:
                if char.isalpha():
                   res =  char.lower()
                else:
                      return ""

          except Exception as inst:
              print "Zeichen nicht ok"
         
          #print "R", type(res)
          return res

      def replaceSpecChars(char):

         #print char, " ", type(char)
            
         #reps = {u'ä':'a', u'ö':'o', u'ü':'u', u'Ä':'A', u'Ö':'O', u'Ü':'U', u'ß':'ss', 'á': 'a', u'é': 'e', u'è': 'e', u'É': 'e', u'È': 'e', u'í': 'i', u'ó': 'o', u'ñ': 'n'}
         #reps = {u'ä':'a', u'ö':'o', u'ü':'u', u'Ä':'A', u'Ö':'O', u'Ü':'U', u'ß':'ss'}
         reps = {u'à': 'a', u'á': 'a', u'è':'e', u'é':'e', u'í':'i', u'ò': u'o', u'ó':'o', u'ñ': 'n'}

         #print reps
         

         for i, j in reps.iteritems():
               char = char.replace(i, j)

         #print "Res ", char
         return char


      c = [None]*len(word2) 
      for i in xrange(len(word2)): 
         c[i] = punct(word2[i].lower())
      res = "".join(c)
      if res.startswith("-"):
         res = res[1:]
      elif res.endswith("-"):
         res = res[0:-1]

      #print res
      res = res.encode('utf-8')
      #print "ferig ", res
      return res


def normalizeWordWithSnowball(word, lang):

      res = ""
      #print word, " ", type(word) # Eingelesenes Wort
      word.decode('utf-8')
      word2 = unicode(word, 'utf-8') # Konvertierung nach unicode
      word3 = word2.encode('utf-8') # Konvertierung nach UTF-8 zum Ausgeben in der Konsole
      #print "T0 ", word3, " ", type(word2)

      # Language-independent normalization:
      # Removing hyphen, lowecasing the word
      def punct(char):

          res = ""
          spec = [u'à', u'á', u'è', u'é', u'í', u'ò', u'ó', u'ñ', u'ä', u'ü', u'ö', u'ß']
     
          try:
             if char in spec or char == "-":
                return char
             else:
                if char.isalpha():
                   res =  char.lower()
                else:
                      return ""

          except Exception as inst:
              print "Zeichen nicht ok"
         
          #print "R", type(res)
          return res

      def replaceSpecChars(char):

         #print char, " ", type(char)
            
         reps = {u'ä':'a', u'ö':'o', u'ü':'u', u'Ä':'A', u'Ö':'O', u'Ü':'U', u'ß':'ss', u'à': 'a', u'á': 'a', u'è':'e', u'é':'e', u'í':'i', u'ò': u'o', u'ó':'o', u'ñ': 'n'}
         #reps = {u'à': 'a', u'á': 'a', u'è':'e', u'é':'e', u'í':'i', u'ò': u'o', u'ó':'o', u'ñ': 'n'}

         #print reps
         if char in reps:
            char = reps[char]

         #print "Res ", char
         return char


      c = [None]*len(word2) 
      for i in xrange(len(word2)): 
         c[i] = replaceSpecChars(punct(word2[i].lower()))
      res = "".join(c)


      try:

         if lang == "de":
            ss = nltk.stem.SnowballStemmer("german")
            res_tmp = ss.stem(res)
         if lang == "en":
            ss = nltk.stem.SnowballStemmer("english")
            res_tmp = ss.stem(res)
         if lang == "fr":
            ss = nltk.stem.SnowballStemmer("french")
            res_tmp = ss.stem(res)
         if lang == "es":
            ss = nltk.stem.SnowballStemmer("spanish")
            res_tmp = ss.stem(res)
            
         res = res_tmp
         res = res.encode('utf-8')
         #print "T1 ", type(res)
         #res = unicode(res, 'latin-1')
         #res.encode()
         #res.encode('ascii')

         #print "Normalisiert: ", word, "-->", res
         return res
         #return res

      except Exception as inst:
         #print "Stemmer hat fehlgeschlagen "
         #print "Normalisiert: ", word, "-->", res
         #print "T2: ", type(res)
         #res = unicode(res, 'latin-1')
         #res.encode('utf-8')
         res = word2
         #return res

      res = res.encode('utf-8')
      return res


# Input: dictionary of word forms with lemma-POS-frequency pairs:
# Output: [['word1', 'word2'], ...]
def identifySimWords(word_lemma_dict, lang):

   if lang == "de":
      min_lev = 0.9
   elif lang == "es":
      min_lev = 0.9
   elif lang == "en":
      min_lev = 0.9
   elif lang == "fr":
      min_lev = 0.9

   res = [] # List of word groups
   mapping_dict = {}
   word_list_tmp = []
   for wf in word_lemma_dict:

      w = word_lemma_dict[wf].wordform
      w_norm = normalizeWordWithSnowball(w, lang)
      #w_norm = normalizeWord(w, lang)
      #print wf, " --> ", w_norm

      if w_norm != '':
         try:
            mapping_dict[w_norm].append(w)
         except KeyError:
            mapping_dict[w_norm] = [w]
            word_list_tmp.append(w_norm)
      else:
         if w != '':
            try:
               mapping_dict[w].append(w)
            except KeyError:
               mapping_dict[w] = [w]
               word_list_tmp.append(w)

   #print mapping_dict
   #word_list = map(lambda k: k, word_lemma_dict.iterkeys())

   word_list = list(set(word_list_tmp))
   word_list.sort()

   #print "Wortliste: " , word_list
   #for w in word_list:
   #   print w


   num_of_comp = 0
   entry_nr = 1
   print "Computing similarity groups..."
   while word_list != []:

      #print entry_nr, "\r",

      start_idx = 0 # First word of a potential group
      #print "Wort ", word_list[start_idx]
      candidates = [(word_list[start_idx], 1)] # List of similarity candidates
      #print "C ", candidates
      if word_list[start_idx][0].isalpha(): # Skip the numbers
         wf_idx = 1
         lev = 1 # Strarting Levenshtein ratio
         while lev >= 0.5:
            if wf_idx < len(word_list):
               w1 = normalizeWordWithSnowball(word_list[start_idx], "de")
               #w1 = normalizeWord(word_list[start_idx], "de")
               w2 = normalizeWordWithSnowball(word_list[wf_idx].lower(), "de")
               #w2 = normalizeWord(word_list[wf_idx].lower(), "de")
               #lev = Levenshtein.ratio(word_list[start_idx].lower(), word_list[wf_idx].lower())
               lev = Levenshtein.ratio(w1, w2)
               #print word_list[start_idx].lower(), ",", word_list[wf_idx].lower(), "=", lev
               candidates.append((word_list[wf_idx], lev))
               wf_idx += 1
            else:
               break
      
      #print "C2", candidates
      sim_group = []
      for (c, l) in candidates:
         if l >= min_lev:
            orig_wfs = mapping_dict[c]
            #print c
            for o in orig_wfs:
               sim_group.append(o)
            word_list.remove(c)
         
      #print "Aehnlichkeitsgruppe = ", sim_group
      #if len(sim_group) >1:
      #   print sim_group
    
      res.append(sim_group)
      
      entry_nr += 1
         
   #print res
   return res


# Input: wf = WordPosLemma object
# and language
def computeNewLemma(lemmas, lang, wf):

   # Creating a dictionary with POS groups:
   # e.g. {'N': ['Ncfsg', 'Np-sn', ...], 'V': [...], ...}
   def createPOSDict(lemmas):

      key_dict = {}  

      for pos in lemmas:
         pos_general = pos[0]
         try:
            key_dict[pos_general].append(pos)
         except KeyError:
            key_dict[pos_general] = [pos]

      return key_dict

   # Gets the most frequent lemma with POS
   # e.g. 'Vx': {'l1': 15, 'l2': 2} --> 'l1' for POS "Vx"
   def getLemma(pos, lemmas, lang, wf):
      
      new_lemma = ""
      freq = 0
      items = [(f, k) for (k, f) in lemmas[pos].items()]
      items.sort()
      items.reverse() # Most frequent form at the beggining of the list
      #print "getLemma items = ", items
      for (f, new) in items:
         #print "new_norm ", normalizeWord(new, lang)
         #print "wf_norm ", normalizeWord(wf, lang)
         if normalizeWord(new, lang) in normalizeWord(wf, lang):
            freq = f
            new_lemma = new
            #print "lemma ok ", new_lemma
            break
      if new_lemma == "":
         (freq, new_lemma) = items[0]
      #print "-->", new_lemma
      return new_lemma, freq

   # Gets the most frequent lemma with the same POS group
   # e.g. {'Vx': {'l1': 15, 'l2': 2}, 'Vy': {'l3': 4, 'l4': 5}, ...}
   # --> "l1 for POS group "V"
   def getMostFreqLemmaForPOSGroup(pos_group, lemmas, lang, wf):

      new_lemma = ""
      lemma_freq = {}

      #print "Suche ", pos_group, " in ", lemmas

      for pos in lemmas:
         if pos[0] == pos_group:

            # German underspecified heads are not considered
            if lang == "de" and pos_group == "N":
               if pos != 'N----':
                  for l in lemmas[pos]:
                     #lemma_items.append((lemmas[pos][l], l))
                     try:
                        lemma_freq[l] += lemmas[pos][l]
                     except KeyError:
                        lemma_freq[l] = lemmas[pos][l]

            else:
               for l in lemmas[pos]:
                  #lemma_items.append((lemmas[pos][l], l))
                  try:
                     lemma_freq[l] += lemmas[pos][l]
                  except KeyError:
                     lemma_freq[l] = lemmas[pos][l]

      # If for DE, there are no lemma candidates,
      # use also the heads
      if lang == "de"  and lemma_freq == {}:
         for pos in lemmas:
            if pos[0] == pos_group:
               for l in lemmas[pos]:
                  #lemma_items.append((lemmas[pos][l], l))
                  try:
                     lemma_freq[l] += lemmas[pos][l]
                  except KeyError:
                     lemma_freq[l] = lemmas[pos][l]

      lemma_items = [(f, l) for (l, f) in lemma_freq.items()]
      #print "lemma_items = ", lemma_items

      lemma_items.sort()
      lemma_items.reverse() # Most frequent form at the beggining of the list
      #new_lemma = lemma_items[0][1]
      #print "MostFreq lemma_items = ", lemma_items
      for (f, new) in lemma_items:
         if normalizeWord(new, lang) in normalizeWord(wf, lang):
            freq = f
            new_lemma = new
            break
      if new_lemma == "":
         new_lemma =lemma_items[0][1]
      #new_lemma = lemma_items[0][1]
      #print "Most freq. gefunden fuer die Gruppe ", pos_group, " ", new_lemma
      
      #return new_lemma
      return new_lemma

   # Search for the new lemma by considering
   # a list of appropriate POS tags
   def getLemmaFromPosAlter(pos_list, lemmas, lang, wf):

      new_lemma = ""
      lemma_cands = {}
      pos_group = pos_list[0][0] # General POS

      # Search for POS
      # When the first pos is found, the 
      # search stops
      for p in pos_list:
         if p in lemmas:
            (new, f) = getLemma(p, lemmas, lang, wf)
            try:
               lemma_cands[new] += f
            except:
               lemma_cands[new] = f
            #print "POS ", p, " gefunden ", new_lemma
            #break

      if lemma_cands != {}:
         #print "B ", lemma_cands.items()
         lemma_items = [(f, l) for (l, f) in lemma_cands.items()]
         lemma_items.sort()
         lemma_items.reverse()
         #print "POSAlter lemma_items = ", lemma_items
         for (f, new) in lemma_items:
            #print "new_norm ", normalizeWord(new, lang)
            #print "wf_norm ", normalizeWord(wf, lang)
            if normalizeWord(new, lang) in normalizeWord(wf, lang):
               freq = f
               new_lemma = new
               break
         if new_lemma == "":
            new_lemma = lemma_items[0][1]
         #new_lemma = lemma_items[0][1]

      # If no appropriate POS is found,
      # the most common lemma of the 
      # POS group is searched
      if new_lemma == "":
         #print "POS nicht gef, suche most frequent"
         new_lemma = getMostFreqLemmaForPOSGroup(pos_group, lemmas, lang, wf)


      #print "neues lemma = ", new_lemma
      return new_lemma


   res_pos_lemma_dict = {}

   # German lemmas
   if lang == "de":

      pos_n = ['Ncfsn', 'Ncfsg', 'Ncfsd', 'Ncfsa', 'Ncmsn', 'Ncmsa', 'Ncnsn', 'Ncnsa']
      pos_v = ['Vmn-----', 'Von-----', 'Van-----', 'Vmip1p--', 'Vmip3p--']

      #print "Suche in ", wf_pos_lemma_dict
      
      pos_dict = createPOSDict(lemmas)
      for p in pos_dict:
         new_lemma = ""
         #print p
         if p not in ['N', 'V']:
            new_lemma = getMostFreqLemmaForPOSGroup(p, lemmas, lang, wf)
            res_pos_lemma_dict[p] = new_lemma
         elif p == "N":
            p_list = pos_dict[p]
            #print "\nSuche Lemma fuer ", wf
            new_lemma = getLemmaFromPosAlter(pos_n, lemmas, lang, wf)
            #new_lemma = getMostFreqLemmaForPOSGroup(p, lemmas, lang)
            res_pos_lemma_dict[p] = new_lemma
            
         elif p == "V":
            p_list = pos_dict[p]
            new_lemma = getLemmaFromPosAlter(pos_v, lemmas, lang, wf)
            #new_lemma = getMostFreqLemmaForPOSGroup(p, lemmas, lang)
            res_pos_lemma_dict[p] = new_lemma   


   elif lang == "es":

      pos_v = ['Vmn-----', 'Von-----', 'Van-----']

      pos_dict = createPOSDict(lemmas)
      for p in pos_dict:
         new_lemma = ""
         if p not in ['V']:
            new_lemma = getMostFreqLemmaForPOSGroup(p, lemmas, lang, wf)
            res_pos_lemma_dict[p] = new_lemma
         elif p == "V":
            p_list = pos_dict[p]
            new_lemma = getLemmaFromPosAlter(pos_v, lemmas, lang, wf)
            res_pos_lemma_dict[p] = new_lemma  
      

   # Other languages
   else:
      
      pos_dict = createPOSDict(lemmas)
      for p in pos_dict:
         new_lemma = getMostFreqLemmaForPOSGroup(p, lemmas, lang, wf)
         res_pos_lemma_dict[p] = new_lemma
               
   return res_pos_lemma_dict


def mergeLemmas(wf_lemma_dict, sim_groups, lang):

   print "Merging lemmas..."

   for sim_group in sim_groups:

      if len(sim_group) > 1:

         new_lemmas = {}
         #print sim_group

         for wf in sim_group:
            wf_obj = wf_lemma_dict[wf]
            for pos in wf_obj.lemmas: # For each POS
               for lemma in wf_obj.lemmas[pos]: # For each lemma with pos POS
                  try:  
                     new_lemmas[pos][lemma] +=  wf_obj.lemmas[pos][lemma]
                  except KeyError:
                     try:
                        new_lemmas[pos][lemma] = wf_obj.lemmas[pos][lemma]
                     except KeyError:
                        new_lemmas[pos] = {}
                        new_lemmas[pos][lemma] = wf_obj.lemmas[pos][lemma]

         # Computing new lemma for the given word form
         # from the merged set of lemmas
         #new_pos_lemmas = computeNewLemma(new_lemmas, lang, wf)
         for wf in sim_group:
            wf_lemma_dict[wf].lemmas = new_lemmas
            new_pos_lemmas = computeNewLemma(new_lemmas, lang, wf)
            wf_lemma_dict[wf].new_lemmas = new_pos_lemmas
         #print "Similarity group = ", sim_group
         #print "Merged lemmas = ", new_lemmas
         #print "New lemmas", new_pos_lemmas, "\n"

      else:
         new_pos_lemmas = computeNewLemma(wf_lemma_dict[sim_group[0]].lemmas, lang, sim_group[0])
         wf_lemma_dict[sim_group[0]].new_lemmas = new_pos_lemmas
         #print "Neue Lemmata B", new_pos_lemmas
         #wf_lemma_dict[sim_group[0]].printObject()

         
   # Update lemmas of the German compounds
   if lang == "de":
      for wf in wf_lemma_dict:
         if wf_lemma_dict[wf].isCompound():
            new_pos_lemmas = wf_lemma_dict[wf].createCompoundLemma()
            if new_pos_lemmas != None:
               wf_lemma_dict[wf].new_lemmas['N'] = new_pos_lemmas
               #"Compound ", wf_lemma_dict[wf].printObject()


   return wf_lemma_dict




def correctLemma(tagged, wf_lemma_dict, lang):

   tagged_file = open(tagged, "r")
   targetEncoding = "utf-8"
   new_tagged = open("tagged.new", "w")
   tagged_line = tagged_file.readline().strip()

   corrected = {}

   while tagged_line:

         # Copying lines with meta data
         if tagged_line.startswith("<"):
            new_tagged.write(tagged_line.strip() + "\n")

         else:

            try:

               split_line = tagged_line.split("\t")
               #print split_line
               
               wf = split_line[0].strip()

               pos = split_line[1].strip()
               p = pos[0] # POS
               morph = pos[1:] # Morphology tags
               
               lemma = split_line[2].strip()

               # Changing lemma only for unknowns;
               # Proper nouns are left unchanged, since
               # they are not used in the extraction process
               #if lemma == '<unknown>':
               if p in ['N', 'V', 'A'] and not pos.startswith("Np"):

                     new_lemma = wf_lemma_dict[wf].new_lemmas[p]
                     if new_lemma == '':
                        new_lemma = wf

                     # All lemmas with exception of German nouns
                     # are lowercased
                     if p != "N" and lang != "de":
                        new_lemma = new_lemma.lower()

                     output_line = wf + "\t" + pos + "\t" + new_lemma # + "\t" + morph
                     
                    #print output_line
                     if new_lemma != lemma:
                        corrected[wf] = (pos, lemma, new_lemma)
                        print wf + "\t" + pos + "\t" + lemma + "\t-->\t" + new_lemma 
                        
                     new_tagged.write(output_line + "\n")

                     #else:
                     #   new_tagged.write(tagged_line.strip() + "\n")

                  #else: 
                  #      new_tagged.write(tagged_line.strip() + "\n")

               else:
                  new_tagged.write(tagged_line.strip() + "\n")
                  

            except Exception as inst:
               print "Error at ", tagged_line.strip(), ": ", inst
               new_tagged.write(tagged_line.strip() + "\n")

         tagged_line = tagged_file.readline()

   for c in corrected:
      print c + "\t" + corrected[c][0] + "\t" + corrected[c][1] + "\t" + corrected[c][2]

   tagged_file.close()
   new_tagged.close()

def correctLemmaSeqRules(wf_lemma_dict, lang, rules):

   def parseRuleFile(rules):

      # Dictionary of rules grouped by POS tag
      # Values are listes of sorted rules, whereby
      # the fist rule is the most restrictive one
      # and the last the most general
      res = {}
      suffixes = [] # List of suffixes being processed

      #print rules

      char_dict = {'voc': ['a', 'e', 'i', 'o', 'u']}
      regexp = re.compile(r'-({?.*})?(.*)')
      regexp2 = re.compile(r'(.+)\[(.*)\]') # Suffix alternatives

      rule_file = open(rules, "r")
      rule_line = rule_file.readline()
      rule_nr = 0
      while rule_line:
         try:

            if rule_line.startswith("#") or rule_line == "\n" or rule_line == "":
               rule_line = rule_file.readline()

            else:
                  #print rule_line
                  splitted_line = rule_line.strip().split("\t")
                  print splitted_line
                  if len(splitted_line) < 3:
                     print "Fehlerhafte Regel-Datei : ", splitted_line
                  else:
                     pos = splitted_line[0].strip()
                     mapping = splitted_line[2].strip()
                     ending = splitted_line[1].strip()
                     end_alters = re.findall(regexp2, ending) #[0][1]
                     #print "end_alters ", end_alters
                     if end_alters != []:
                        ctx = end_alters[0][0]
                        endings = end_alters[0][1].split("|")
                        for e in endings:
                           suff = ctx.strip() + e.strip()
                           suffixes.append(suff[1:])
                           if pos not in res:         
                              res[pos] = [(suff, mapping)]
                           else:
                              res[pos].append((suff, mapping))
                     else:
                        suff = re.findall(regexp, ending)[0][1]
                        suffixes.append(suff)
                        if pos not in res:         
                           res[pos] = [(ending, mapping)]
                        else:
                           res[pos].append((ending, mapping))

                     
                     #if pos not in res:         
                      #  res[pos] = [(ending, mapping)]
                     #else:
                     #   res[pos].append((ending, mapping))
                     

                  rule_nr += 1
                  rule_line = rule_file.readline()

         except Exception as inst:
            print "Fehler: ", inst, rule_line
            rule_line = rule_file.readline()

      rule_file.close()
      #print rule_nr, " : ", res, "\n", suffixes
      return res, suffixes

  
   def correct(old_lemma, pos, rule_dict):

      new_lemma = old_lemma
      lemma_hyp = ""
      regexp = re.compile(r'-({?.*})?(.*)')
      vocals = ['a', 'e', 'i', 'o', 'u']
      replacements =[]
      not_confirmed_replacements = []

      #print "Kandidat: ", old_lemma, " ", pos, " ", suffix 

      if pos in rule_dict:

         for (constr, mapping) in rule_dict[pos]:

            #print "Regel: ", constr, "-->", mapping
            constr_parts = re.findall(regexp, constr)[0]
            #print "constr_parts ", constr_parts
            mapping_parts = re.findall(regexp, mapping)[0]
            #print "mapping_parts ", mapping_parts

            #suff_len = 0

            ctx = constr_parts[0]
            ending = constr_parts[1]

            if old_lemma.endswith(ending): # (constr_parts[1]):
               # Apply rule
               #print "Regel gefunden: ", constr, "-->", mapping
               suff_len = len(ending) #constr_parts[1])
               #print "H: ", old_lemma[len(old_lemma)-suff_len-1]

               if ctx != '':
                  if "{doublecons}" in ctx:
                     if old_lemma[len(old_lemma)-suff_len-1] not in vocals and old_lemma[len(old_lemma)-suff_len-1] == old_lemma[len(old_lemma)-suff_len-2]:
                        suff_len += 2
                     else: 
                        #print "Doch nicht"
                        continue
                  elif "{cons}" in ctx:
                     if old_lemma[len(old_lemma)-suff_len-1] not in vocals:
                        suff_len += 1
                     else:
                        #print "Doch nicht"
                        continue
                  elif "{voc}" in ctx:
                     if old_lemma[len(old_lemma)-suff_len-1] in vocals:
                        suff_len += 1
                     else:
                        #print "Doch nicht"
                        continue

               #print "suff_len = ", suff_len
               suff_idx = ((len(old_lemma) - suff_len), len(old_lemma))
               #print "suff_idx = ", suff_idx
                  
               i = 0
               new_lemma = new_lemma[0:suff_idx[0]] # Part before suffix
               #print "new_lemma = ", new_lemma
               tmp_suff = old_lemma[suff_idx[0]:suff_idx[1]]
               #print "1: tmp_suff = ", tmp_suff
               if ctx != '' and mapping_parts[0] != '':
                  if "{doublecons}" in constr_parts[0] and "{cons}" in mapping_parts[0]:
                     double_cons = tmp_suff[0:2]
                     tmp_suff = tmp_suff.replace(double_cons, double_cons[0])
                     #print "2: tmp_suff = ", tmp_suff
                  elif "{cons}" in constr_parts[0] and "{cons}" in mapping_parts[0]:
                     tmp_suff = tmp_suff.replace(constr_parts[1].strip(), mapping_parts[1].strip())
               elif ctx != '':
                  if "{voc}" in constr_parts[0]:
                     if tmp_suff[0] in vocals:
                         tmp_suff = tmp_suff.replace(constr_parts[1].strip(), mapping_parts[1].strip())
               else: # no context
                   tmp_suff = tmp_suff.replace(ending, mapping_parts[1].strip())

               #print "replace ", ending, " by ", mapping_parts[1]
               tmp_suff = tmp_suff.replace(constr_parts[1].strip(), mapping_parts[1].strip())
               #print "3: tmp_suff = ", tmp_suff
               new_lemma += tmp_suff
               #print "final ", new_lemma
               #if pos != "N":
               #  new_lemma = new_lemma.lower()
               break
            

      #print "Check ", old_lemma, "/", pos, " --> ", new_lemma
      confirmed = False
      if old_lemma != new_lemma:
         if new_lemma not in wf_lemma_dict:
            #print "Nicht bestaetigt: ", old_lemma, " --> ", new_lemma
            lemma_hyp = new_lemma
            new_lemma = lemma_hyp
         else:
            #print "Bestaetigt ", old_lemma, " --> ", new_lemma
            confirmed = True
            #break
      else:
         confirmed = None

      return (new_lemma, confirmed, lemma_hyp)

   def printTupels(tupel_list1, tupel_list2):

      log_file = open("lemma_corr.log", "w")

      def writeData(tupel_list):
         i = 0
         log_file.write("-----\n")
         for t in tupel_list:
            log_file.write(t[0] + ", " + t[1])
            if i < 4:
               log_file.write("\t")
               i += 1
            else:
               log_file.write("\n")
               i = 0
         log_file.write("\n-----\n\n")

      writeData(tupel_list1)
      writeData(tupel_list2)

      log_file.close()
               


   res = {}

   (rule_dict, suffixes) = parseRuleFile(rules)
   #print "Dict: ", rule_dict
   #print "Suffixes: ", suffixes

   replacements = []
   not_confirmed_replacements = []

   for wf in wf_lemma_dict:

      #print wf, " ", wf_lemma_dict[wf].new_lemmas
      #to_correct = False

      try:

         if wf_lemma_dict[wf].new_lemmas != {}:

            # For each POS, try to correct the hypothesized lemma
            for pos in wf_lemma_dict[wf].new_lemmas:
               to_correct = False
               if pos in rule_dict:
                  #print "Zu korr. ", wf_lemma_dict[wf].new_lemmas[pos]
                  lemma = wf_lemma_dict[wf].new_lemmas[pos]
                  
                  #to_correct = False
                  suffix = ""
                  for s in suffixes:
                     #print s,  " ",  lemma
                     if lemma.endswith(s):
                        to_correct = True
                        suffix = s
                        break
                  if to_correct:
                     new_lemma = ""
                     #print "Korrektur ", wf_lemma_dict[wf].new_lemmas[pos]
                     (new_lemma, conf, lemma_hyp) = correct(lemma, pos, rule_dict)
                     #print "Korrektur ", wf_lemma_dict[wf].new_lemmas[pos], " --> ", new_lemma
                     if conf:
                        replacements.append((lemma, new_lemma))
                        wf_lemma_dict[wf].new_lemmas[pos] = new_lemma
                        
                     else:
                        not_confirmed_replacements.append((lemma, lemma_hyp))
                        wf_lemma_dict[wf].new_lemmas[pos] = new_lemma
                  #else:
                  #   new_lemma = lemma
                  #   conf = None

                  #if conf != None:
                  #   if conf:
                  #      replacements.append((lemma, new_lemma))
                  #      wf_lemma_dict[wf].lemmas[pos] = new_lemma
                  #   else:
                  #      not_confirmed_replacements.append((lemma, lemma_hyp))
                  #      wf_lemma_dict[wf].lemmas[pos] = new_lemma

            
         
      except Exception as inst:
         print "Fehler bei ", wf, " : ", inst
         pass



   #print "Nicht verif. Korrekturen: ", not_confirmed_replacements
   #printTupels(set(not_confirmed_replacements), set(replacements))

# Input: dictionary with lemmas computed by 
# similarity-based approach and the
# tagged specialized file
# If POS in rules, the rules are applied, otherwise
# the lemma is taken from the dictionary
def correctLemmaSeqRules2(tagged, wf_lemma_dict_sim, lang, rules):

   def parseRuleFile(rules):

      # Dictionary of rules grouped by POS tag
      # Values are listes of sorted rules, whereby
      # the fist rule is the most restrictive one
      # and the last the most general
      res = {}
      suffixes = [] # List of suffixes being processed

      #print rules

      char_dict = {'voc': ['a', 'e', 'i', 'o', 'u']}
      regexp = re.compile(r'-({?.*})?(.*)')
      regexp2 = re.compile(r'(.+)\[(.*)\]') # Suffix alternatives

      rule_file = open(rules, "r")
      rule_line = rule_file.readline()
      rule_nr = 0
      while rule_line:
         try:

            if rule_line.startswith("#") or rule_line == "\n" or rule_line == "":
               rule_line = rule_file.readline()

            else:
                  #print rule_line
                  splitted_line = rule_line.strip().split("\t")
                  #print splitted_line
                  if len(splitted_line) < 3:
                     print "Erroneous rule file: ", splitted_line
                  else:
                     pos = splitted_line[0].strip()
                     mapping = splitted_line[2].strip()
                     ending = splitted_line[1].strip()
                     end_alters = re.findall(regexp2, ending) #[0][1]
                     #print "end_alters ", end_alters
                     if end_alters != []:
                        ctx = end_alters[0][0]
                        endings = end_alters[0][1].split("|")
                        for e in endings:
                           suff = ctx.strip() + e.strip()
                           suffixes.append(suff[1:])
                           if pos not in res:         
                              res[pos] = [(suff, mapping)]
                           else:
                              res[pos].append((suff, mapping))
                     else:
                        suff = re.findall(regexp, ending)[0][1]
                        suffixes.append(suff)
                        if pos not in res:         
                           res[pos] = [(ending, mapping)]
                        else:
                           res[pos].append((ending, mapping))

                     
                     #if pos not in res:         
                      #  res[pos] = [(ending, mapping)]
                     #else:
                     #   res[pos].append((ending, mapping))
                     

                  rule_nr += 1
                  rule_line = rule_file.readline()

         except Exception as inst:
            print "Error: ", inst, rule_line
            rule_line = rule_file.readline()

      rule_file.close()
      #print rule_nr, " : ", res, "\n", suffixes
      return res, suffixes

   def correct(old_lemma, pos, rule_dict, wf_lemma_dict):

      new_lemma = old_lemma
      lemma_hyp = ""
      regexp = re.compile(r'-({?.*})?(.*)')
      vocals = ['a', 'e', 'i', 'o', 'u']
      replacements =[]
      not_confirmed_replacements = []
      pos_group = pos[0]
      confirmed = None

      #print "Kandidat: ", old_lemma, " ", pos, " ", suffix 

      if pos_group in rule_dict:


         for (constr, mapping) in rule_dict[pos_group]:

            #print "Regel: ", constr, "-->", mapping
            constr_parts = re.findall(regexp, constr)[0]
            #print "constr_parts ", constr_parts
            mapping_parts = re.findall(regexp, mapping)[0]
            #print "mapping_parts ", mapping_parts

            #suff_len = 0

            ctx = constr_parts[0]
            ending = constr_parts[1]

            if old_lemma.endswith(ending): # (constr_parts[1]):
               # Apply rule
               #print "Regel gefunden: ", constr, "-->", mapping
               suff_len = len(ending) #constr_parts[1])
               #print "suff_len ", suff_len
               before_suff = old_lemma[len(old_lemma)-suff_len-1]
               #print "before_suff: ", before_suff

               ctx_ok = True

               if ctx != '':
                  if "{doublecons}" in ctx:
                     if before_suff not in vocals and before_suff == old_lemma[len(old_lemma)-suff_len-2]:
                        suff_len += 2
                     else: 
                        #print "1 Doch nicht"
                        ctx_ok = False
                  elif "{cons}" in ctx:
                     if before_suff not in vocals:
                        suff_len += 1
                     else:
                        #print "2 Doch nicht"
                        ctx_ok = False
                  elif "{voc}" in ctx:
                     if before_suff in vocals:
                        suff_len += 1
                     else:
                        #print "3 Doch nicht"
                        ctx_ok = False

               if ctx_ok:

                  #print "suff_len = ", suff_len
                  suff_idx = ((len(old_lemma) - suff_len), len(old_lemma))
                  #print "suff_idx = ", suff_idx
                  
                  i = 0
                  new_lemma = new_lemma[0:suff_idx[0]] # Part before suffix
                  #print "new_lemma = ", new_lemma
                  tmp_suff = old_lemma[suff_idx[0]:suff_idx[1]]
                  #print "1: tmp_suff = ", tmp_suff
                  if ctx != '' and mapping_parts[0] != '':
                     if "{doublecons}" in constr_parts[0] and "{cons}" in mapping_parts[0]:
                        double_cons = tmp_suff[0:2]
                        tmp_suff = tmp_suff.replace(double_cons, double_cons[0])
                        #print "2: tmp_suff = ", tmp_suff
                     elif "{cons}" in constr_parts[0] and "{cons}" in mapping_parts[0]:
                        tmp_suff = tmp_suff.replace(constr_parts[1].strip(), mapping_parts[1].strip())
                  elif ctx != '':
                     if "{voc}" in constr_parts[0]:
                        if tmp_suff[0] in vocals:
                           tmp_suff = tmp_suff.replace(constr_parts[1].strip(), mapping_parts[1].strip())
                  else: # no context
                     tmp_suff = tmp_suff.replace(ending, mapping_parts[1].strip())

                  #print "replace ", ending, " by ", mapping_parts[1]
                  tmp_suff = tmp_suff.replace(constr_parts[1].strip(), mapping_parts[1].strip())
                  #print "3: tmp_suff = ", tmp_suff
                  new_lemma += tmp_suff
                  #print "final ", new_lemma
                  #if pos != "N":
                  #  new_lemma = new_lemma.lower()
                  #break
            

               #print "Check ", old_lemma, "/", pos, " --> ", new_lemma
               confirmed = False
               if old_lemma != new_lemma:
                  if new_lemma not in wf_lemma_dict[pos_group]:
                     #print "Nicht bestaetigt: ", old_lemma, " --> ", new_lemma
                     lemma_hyp = new_lemma
                     new_lemma = lemma_hyp
                  else:
                     #print "Bestaetigt: ", old_lemma, " --> ", new_lemma
                     confirmed = True
                     break
               else:
               #print "Keine Aenderung: ", new_lemma
                  confirmed = None
                  break

      #print confirmed, ": ", old_lemma, "/", pos, " --> ", new_lemma
      return (new_lemma, confirmed, lemma_hyp)

   def printTupels(tupel_list1, tupel_list2):

      log_file = open("lemma_corr.log", "w")

      def writeData(tupel_list):
         i = 0
         log_file.write("-----\n")
         for t in tupel_list:
            log_file.write(t[0] + ", " + t[1])
            if i < 4:
               log_file.write("\t")
               i += 1
            else:
               log_file.write("\n")
               i = 0
         log_file.write("\n-----\n\n")

      writeData(tupel_list1)
      writeData(tupel_list2)

      log_file.close()
               


   print "Computing lemmas with rules..."

   #print wf_lemma_dict_sim

   replacements = []
   not_confirmed_replacements = []
   wf_lemma_dict = {}
   (rule_dict, suffixes) = parseRuleFile(rules)

   tagged_file = open(tagged, "r")
   out_file = open("tagged.new", "w")

   # Creating a pos-word-lemma dictionary:
   # {POS: {'word1':'', 'lemma1':'', ...}, ...}
   tagged_line = tagged_file.readline()

   while tagged_line:
      
      if not tagged_line.startswith("<"):
         try:
            (wf, pos, lemma) = tagged_line.strip().split("\t")
            pos_group = pos[0]
            try:
               wf_lemma_dict[pos_group][wf.strip()] = ""
               wf_lemma_dict[pos_group][lemma.strip()] = ""
            except KeyError as inst:
               wf_lemma_dict[pos_group] = {}
               wf_lemma_dict[pos_group][wf.strip()] = ""
               wf_lemma_dict[pos_group][lemma.strip()] = ""
         except Exception as inst:
            pass
      tagged_line = tagged_file.readline()

   tagged_file.close()
   
   # Correcting lemmas
   tagged_file = open(tagged, "r")
   tagged_line = tagged_file.readline()
   while tagged_line:
      
      if tagged_line.startswith("<"):
         out_file.write(tagged_line)
         
      else:

         try:

            (wf, pos, old_lemma) = tagged_line.strip().split("\t")
            if old_lemma == "<unknown>":
               old_lemma = wf
            #old_lemma_norm = old_lemma
            old_lemma_norm = normalizeWordForRules(old_lemma, lang)
            to_correct = False
            new_lemma = ""
            pos_group = pos[0]
            if (not pos.startswith("Np")) and pos_group in rule_dict:
               #print "Zu korr. ", old_lemma
               #old_lemma_norm = normalizeWordForRules(old_lemma, lang)
               #print "Norm ", old_lemma_norm
               suffix = ""
               for s in suffixes:
                  #print s,  " ",  lemma
                  if old_lemma_norm.endswith(s):
                     to_correct = True
                     suffix = s
                     break
               if to_correct:
                  new_lemma = ""
                  #print "Korrektur von ", old_lemma
                  (new_lemma_hyp, conf, lemma_hyp) = correct(old_lemma_norm, pos, rule_dict, wf_lemma_dict)
                  #print conf, ": ", wf, "/", pos, " --> ", new_lemma
            
                  if conf:
                     replacements.append((old_lemma, new_lemma))
                     new_lemma = new_lemma_hyp
                  else:
                     not_confirmed_replacements.append((old_lemma, new_lemma))
                     if lang != "en":
                        new_lemma = lemma_hyp

            # Other word categories
            else:
               #print "Nicht zu korrigieren: ", wf
               try:
                  new_lemma = wf_lemma_dict_sim[wf].new_lemmas[pos_group]
                  #print "Sim_lemma = ", new_lemma
               except KeyError as inst:
                  new_lemma = old_lemma_norm

            if new_lemma == "":
               #print "Leeres lemma fuer ", wf
               try:
                  new_lemma = wf_lemma_dict_sim[wf].new_lemmas[pos_group]
                  #print "Sim_lemma = ", new_lemma
                  if new_lemma == "":
                     if old_lemma_norm != '': # Deleted special characters
                        new_lemma = old_lemma_norm
                     else:
                        new_lemma = old_lemma
               except KeyError as inst:
                  if old_lemma_norm != '': # Deleted special characters
                     new_lemma = old_lemma_norm
                  else:
                     new_lemma = old_lemma
            if pos.startswith("N"):
               if lang != "de":
                  new_lemma = new_lemma.lower()
               else:
                  new_lemma = new_lemma[0].upper() + new_lemma[1:]
            else:
               new_lemma = new_lemma.lower()

            out_file.write(wf + "\t" + pos + "\t" + new_lemma + "\n")

         except Exception as inst:
            print "Fehler bei ", tagged_line.strip(), " : ", inst
            out_file.write(tagged_line)

      tagged_line = tagged_file.readline()


   tagged_file.close()
   out_file.close()


# Direct correction of an input word using rules
def correctLemmaSeqRulesDirectly(tagged, lang, rules):

   print "Applying rule-based lemma correction"

   def parseRuleFile(rules):

      # Dictionary of rules grouped by POS tag
      # Values are listes of sorted rules, whereby
      # the fist rule is the most restrictive one
      # and the last the most general
      res = {}
      suffixes = [] # List of suffixes being processed

      #print rules

      char_dict = {'voc': ['a', 'e', 'i', 'o', 'u']}
      regexp = re.compile(r'-({?.*})?(.*)')
      regexp2 = re.compile(r'(.+)\[(.*)\]') # Suffix alternatives

      rule_file = open(rules, "r")
      rule_line = rule_file.readline()
      rule_nr = 0
      while rule_line:
         try:

            if rule_line.startswith("#") or rule_line == "\n" or rule_line == "":
               rule_line = rule_file.readline()

            else:
                  #print rule_line
                  splitted_line = rule_line.strip().split("\t")
                  #print splitted_line
                  if len(splitted_line) < 3:
                     print "Erroneous rule file: ", splitted_line
                  else:
                     pos = splitted_line[0].strip()
                     mapping = splitted_line[2].strip()
                     ending = splitted_line[1].strip()
                     end_alters = re.findall(regexp2, ending) #[0][1]
                     #print "end_alters ", end_alters
                     if end_alters != []:
                        ctx = end_alters[0][0]
                        endings = end_alters[0][1].split("|")
                        for e in endings:
                           suff = ctx.strip() + e.strip()
                           suffixes.append(suff[1:])
                           if pos not in res:         
                              res[pos] = [(suff, mapping)]
                           else:
                              res[pos].append((suff, mapping))
                     else:
                        suff = re.findall(regexp, ending)[0][1]
                        suffixes.append(suff)
                        if pos not in res:         
                           res[pos] = [(ending, mapping)]
                        else:
                           res[pos].append((ending, mapping))

                     
                     #if pos not in res:         
                      #  res[pos] = [(ending, mapping)]
                     #else:
                     #   res[pos].append((ending, mapping))
                     

                  rule_nr += 1
                  rule_line = rule_file.readline()

         except Exception as inst:
            print "Fehler: ", inst, rule_line
            rule_line = rule_file.readline()

      rule_file.close()
      #print rule_nr, " : ", res, "\n", suffixes
      return res, suffixes

   def correct(old_lemma, pos, rule_dict, wf_lemma_dict):

      new_lemma = old_lemma
      lemma_hyp = ""
      regexp = re.compile(r'-({?.*})?(.*)')
      vocals = ['a', 'e', 'i', 'o', 'u']
      replacements =[]
      not_confirmed_replacements = []
      pos_group = pos[0]
      confirmed = None

      #print "Kandidat: ", old_lemma, " ", pos, " ", suffix 

      if pos_group in rule_dict:


         for (constr, mapping) in rule_dict[pos_group]:

            #print "Regel: ", constr, "-->", mapping
            constr_parts = re.findall(regexp, constr)[0]
            #print "constr_parts ", constr_parts
            mapping_parts = re.findall(regexp, mapping)[0]
            #print "mapping_parts ", mapping_parts

            #suff_len = 0

            ctx = constr_parts[0]
            ending = constr_parts[1]

            if old_lemma.endswith(ending): # (constr_parts[1]):
               # Apply rule
               #print "Regel gefunden: ", constr, "-->", mapping
               suff_len = len(ending) #constr_parts[1])
               #print "suff_len ", suff_len
               before_suff = old_lemma[len(old_lemma)-suff_len-1]
               #print "before_suff: ", before_suff

               ctx_ok = True

               if ctx != '':
                  if "{doublecons}" in ctx:
                     if before_suff not in vocals and before_suff == old_lemma[len(old_lemma)-suff_len-2]:
                        suff_len += 2
                     else: 
                        #print "1 Doch nicht"
                        ctx_ok = False
                  elif "{cons}" in ctx:
                     if before_suff not in vocals:
                        suff_len += 1
                     else:
                        #print "2 Doch nicht"
                        ctx_ok = False
                  elif "{voc}" in ctx:
                     if before_suff in vocals:
                        suff_len += 1
                     else:
                        #print "3 Doch nicht"
                        ctx_ok = False

               if ctx_ok:

                  #print "suff_len = ", suff_len
                  suff_idx = ((len(old_lemma) - suff_len), len(old_lemma))
                  #print "suff_idx = ", suff_idx
                  
                  i = 0
                  new_lemma = new_lemma[0:suff_idx[0]] # Part before suffix
                  #print "new_lemma = ", new_lemma
                  tmp_suff = old_lemma[suff_idx[0]:suff_idx[1]]
                  #print "1: tmp_suff = ", tmp_suff
                  if ctx != '' and mapping_parts[0] != '':
                     if "{doublecons}" in constr_parts[0] and "{cons}" in mapping_parts[0]:
                        double_cons = tmp_suff[0:2]
                        tmp_suff = tmp_suff.replace(double_cons, double_cons[0])
                        #print "2: tmp_suff = ", tmp_suff
                     elif "{cons}" in constr_parts[0] and "{cons}" in mapping_parts[0]:
                        tmp_suff = tmp_suff.replace(constr_parts[1].strip(), mapping_parts[1].strip())
                  elif ctx != '':
                     if "{voc}" in constr_parts[0]:
                        if tmp_suff[0] in vocals:
                           tmp_suff = tmp_suff.replace(constr_parts[1].strip(), mapping_parts[1].strip())
                  else: # no context
                     tmp_suff = tmp_suff.replace(ending, mapping_parts[1].strip())

                  #print "replace ", ending, " by ", mapping_parts[1]
                  tmp_suff = tmp_suff.replace(constr_parts[1].strip(), mapping_parts[1].strip())
                  #print "3: tmp_suff = ", tmp_suff
                  new_lemma += tmp_suff
                  #print "final ", new_lemma
                  #if pos != "N":
                  #  new_lemma = new_lemma.lower()
                  #break
            

               #print "Check ", old_lemma, "/", pos, " --> ", new_lemma
               confirmed = False
               if old_lemma != new_lemma:
                  if new_lemma not in wf_lemma_dict[pos_group]:
                     #print "Nicht bestaetigt: ", old_lemma, " --> ", new_lemma
                     lemma_hyp = new_lemma
                     new_lemma = lemma_hyp
                  else:
                  #print "Bestaetigt: ", old_lemma, " --> ", new_lemma
                     confirmed = True
                     break
               else:
               #print "Keine Aenderung: ", new_lemma
                  confirmed = None
                  break

      #print confirmed, ": ", old_lemma, "/", pos, " --> ", new_lemma
      return (new_lemma, confirmed, lemma_hyp)

   def printTupels(tupel_list1, tupel_list2):

      log_file = open("lemma_corr.log", "w")

      def writeData(tupel_list):
         i = 0
         log_file.write("-----\n")
         for t in tupel_list:
            log_file.write(t[0] + ", " + t[1])
            if i < 4:
               log_file.write("\t")
               i += 1
            else:
               log_file.write("\n")
               i = 0
         log_file.write("\n-----\n\n")

      writeData(tupel_list1)
      writeData(tupel_list2)

      log_file.close()
               


   res = {}

   (rule_dict, suffixes) = parseRuleFile(rules)
   #print "Dict: ", rule_dict
   #print "Suffixes: ", suffixes

   replacements = []
   not_confirmed_replacements = []
   wf_lemma_dict = {}

   tagged_file = open(tagged, "r")
   out_file = open("tagged.new", "w")

   # Creating a pos-word-lemma dictionary:
   # {POS: {'word1':'', 'lemma1':'', ...}, ...}
   tagged_line = tagged_file.readline()

   while tagged_line:
      
      if not tagged_line.startswith("<"):
         try:
            (wf, pos, lemma) = tagged_line.strip().split("\t")
            pos_group = pos[0]
            try:
               wf_lemma_dict[pos_group][wf.strip()] = ""
               wf_lemma_dict[pos_group][lemma.strip()] = ""
            except KeyError as inst:
               wf_lemma_dict[pos_group] = {}
               wf_lemma_dict[pos_group][wf.strip()] = ""
               wf_lemma_dict[pos_group][lemma.strip()] = ""
         except Exception as inst:
            pass
      tagged_line = tagged_file.readline()

   tagged_file.close()
   
   # Correcting lemmas
   tagged_file = open(tagged, "r")
   tagged_line = tagged_file.readline()
   while tagged_line:
      
      if tagged_line.startswith("<"):
         out_file.write(tagged_line)
         
      else:

         try:

            (wf, pos, old_lemma) = tagged_line.strip().split("\t")
            if old_lemma == "<unknown>":
               old_lemma = wf
            #old_lemma_norm = old_lemma
            old_lemma_norm = normalizeWordForRules(old_lemma, lang)
            to_correct = False
            new_lemma = ""
            pos_group = pos[0]
            if (not pos.startswith("Np")) and pos_group in rule_dict:
               #print "Zu korr. ", old_lemma
               #old_lemma_norm = normalizeWordForRules(old_lemma, lang)
               #print "Norm ", old_lemma_norm
               suffix = ""
               for s in suffixes:
                  #print s,  " ",  lemma
                  if old_lemma_norm.endswith(s):
                     to_correct = True
                     suffix = s
                     break
               if to_correct:
                  new_lemma = ""
                  #print "Korrektur von ", old_lemma
                  (new_lemma_hyp, conf, lemma_hyp) = correct(old_lemma_norm, pos, rule_dict, wf_lemma_dict)
                  #print conf, ": ", wf, "/", pos, " --> ", new_lemma
            
                  if conf:
                     replacements.append((old_lemma, new_lemma))
                     new_lemma = new_lemma_hyp
                  else:
                     not_confirmed_replacements.append((old_lemma, new_lemma))

            # Other word categories
            else:
               new_lemma = old_lemma

            if new_lemma == "":
               if old_lemma_norm != '': # Deleted special characters
                  new_lemma = old_lemma_norm
               else:
                  new_lemma = old_lemma
            new_lemma = new_lemma.lower()

            if pos.startswith("N"):
               if lang != "de":
                  new_lemma = new_lemma.lower()
               else:
                  new_lemma = new_lemma[0].upper() + new_lemma[1:]
            else:
               new_lemma = new_lemma.lower()

            out_file.write(wf + "\t" + pos + "\t" + new_lemma + "\n")

         except Exception as inst:
            print "Fehler bei ", tagged_line.strip(), " : ", inst
            out_file.write(tagged_line)

      tagged_line = tagged_file.readline()


   tagged_file.close()
   out_file.close()

   #print "Nicht verif. Korrekturen: ", not_confirmed_replacements
   #for (i, j) in not_confirmed_replacements:
   #   if i != j:
   #      print i, " --> ", j
   #printTupels(set(not_confirmed_replacements), set(replacements))
   #return new_lemma

def correctLemmaWithRules(tagged, lang, rules):

   tagged_file = open(tagged, "r")
   out_file = open("tagged.new", "w")

   tagged_line = tagged_file.readline()

   while tagged_line:
      
      if tagged_line.startswith("<"):
         out_file.write(tagged_line)
         
      else:

         try:
            (wf, pos, lemma) = tagged_line.strip().split("\t")
            new_lemma = lemma
            new_lemma = correctLemmaSeqRulesDirectly(wf, pos, lemma, lang, rules)

            if new_lemma == "":
               new_lemma = lemma

            out_file.write(wf + "\t" + pos + "\t" + new_lemma + "\n")

         except Exception as inst:
            print "Fehler bei ", tagged_line.strip(), " : ", inst
            out_file.write(tagged_line)

      tagged_line = tagged_file.readline()

   tagged_file.close()
   out_file.close()


################# MAIN #########################

def main():

   getOpts(sys.argv[1:])

   if _lang == "de" or _lang == "fr":
      wf_lemma_dict = {}

      wf_lemma_dict = getLemmaWordListWithPOSFromTagged(_SPECTagged, wf_lemma_dict)
      if _lang == "de":
         wf_lemma_dict = getWordsFromSplit(_split, wf_lemma_dict)
      

      sim_groups = identifySimWords(wf_lemma_dict, _lang)

      wf_lemma_dict = mergeLemmas(wf_lemma_dict, sim_groups, _lang)
      correctLemmaSeqRules2(_SPECTagged, wf_lemma_dict, _lang, _rules)
      

   elif _lang == "es" or _lang == "en" or _lang == "fr":

      correctLemmaSeqRulesDirectly(_SPECTagged, _lang, _rules)

 
###############################################

#profile.run('main()')
main()
