#! /usr/bin/python
# -*- coding: utf-8 -*-


import sys, os, getopt

#####################################################
# GLOBAL VARIABLES

input = ""                          #  RFTagger/TreeTagger output
lang = ""                           # Language
tagger = ""                       # Tagger used for DE: RFTagger (rf) or TreeTagger (tt)
unknown = ""                   # Replace or retain unknown lemmas
out = "corpus.multag"       # Output file

#####################################################
# Getting programme options
def getOpts(argv):

   def usage():
      print "Usage: python ", sys.argv[0], " -i tagged-input-file -l language -t tt or rf -u yes or no"

   opts, args = getopt.getopt(argv, "hi:l:t:u:", ["--help", "tagged-file=", "language=", "tagger=", "--unknown="])
   #print opts
   for opt, arg in opts:
        #print opt, arg
        if opt in ("-h", "--help"):
            usage()                     
            sys.exit()                  
        elif opt in ('-i', "--tagged-file"):
            global rft_input          
            rft_input = arg   
        elif opt in ('-l', "--language"):
            global lang          
            lang = arg
        elif opt in ('-t', "--tagger"):
           global tagger
           tagger = arg
        elif opt in ('-u', "--unknown"):
           global unknown
           unknown = arg

   print "Language: ", lang
   print "Tagger: ", tagger
   if lang in ['en', 'es', 'fr']:
      print "Warning: only POS tags marking nouns, adjectives and verbs\nwill be converted into MulText!"
               

def retagGermanRF(morphpos):

   def posToMulText(morphpos):
      
      res = ""
      
      mapValDict = {}
      mapValDict['N'] = {'Reg': 'c', 'Name': 'p', 'Nom': 'n', 'Gen': 'g', 'Dat': 'd', 'Acc': 'a', 'Masc': 'm', 'Fem': 'f', 'Neut': 'n', 'Sg': 's', 'Pl': 'p'}
      mapValDict['ADJA'] = {'Pos': 'p', 'Comp': 'c', 'Sup': 's', 'Nom': 'n', 'Gen': 'g', 'Dat': 'd', 'Acc': 'a', 'Masc': 'm', 'Fem': 'f', 'Neut': 'n', 'Sg': 's', 'Pl': 'p'}
      mapValDict['ART'] = {'Def': 'd', 'Indef' : 'i', 'Nom': 'n', 'Gen': 'g', 'Dat': 'd', 'Acc': 'a', 'Masc': 'm', 'Fem': 'f', 'Neut': 'n', 'Sg': 's', 'Pl': 'p'}
      mapValDict['VFIN'] = {'Aux': 'a', 'Full': 'm', 'Mod': 'o', '1': '1', '2': '2', '3': '3', 'Sg': 's', 'Pl': 'p', 'Pres': 'p', 'Ind': 'i', 'Past': 's', 'Subj': 's'}
      mapValDict['VINF'] = {'Aux': 'a', 'Full': 'm', 'Mod': 'o', '1': '1', '2': '2', '3': '3', 'Sg': 's', 'Pl': 'p', 'Pres': 'p', 'Ind': 'i', 'Past': 's', 'Subj': 's', 'zu': 'u'}
      mapValDict['PRO'] = {'Dem': 'd', 'Inter': 't', 'Rel': 'r', 'Indef': 'i', 'Poss': 'p', '1': '1', '2': '2', '3': '3', 'Nom': 'n', 'Gen': 'g', 'Dat': 'd', 'Acc': 'a', 'Masc': 'm', 'Fem': 'f', 'Neut': 'n', 'Sg': 's', 'Pl': 'p', 'Refl': 'x', 'Pers': 'p', '*': '-'}
      mapValDict['CONJ'] = {'SubInf': 'i', 'SubFin': 's', 'Comp': 'v', 'Coord': 'c'}
      mapValDict['PART'] = {'Verb': 'v', 'Deg': 's', 'Ans': '-', 'Zu': 'i', 'Neg': 'n'}
      mapPosDict = {'SYM': 'S', 'APPR': 'Sp-', 'APPRART': 'Sp-', 'ADJD': 'R--', 'ADV': 'R--', 'CARD': 'X', 'FM': 'X', 'PROADV': 'R--', 'ITJ': 'I', 'APPO': 'St-', 'APZR': 'St-'}

      morphpos_splitted = morphpos.split(".")
      #print morphpos_splitted
      pos = morphpos_splitted[0]
      #print "Pos = ", pos
      res = pos
      if pos in mapPosDict:
         pos_new = mapPosDict[pos]
         res = pos_new
         #print "Alt: ", pos, "; Neu: ", pos_new

      elif pos in mapValDict:
         pos_new = ""

         if pos == "N":
            new_pos = "N" # Pos
            new_pos += mapValDict['N'][morphpos_splitted[1]]  # Type
            if morphpos_splitted[4] != "*":
               new_pos += mapValDict['N'][morphpos_splitted[4]] # Gender
            else:
               new_pos += "-"
            if morphpos_splitted[3] != "*":
               new_pos += mapValDict['N'][morphpos_splitted[3]] # Number
            else:
               new_pos += "-"
            if morphpos_splitted[2] != "*":
               new_pos += mapValDict['N'][morphpos_splitted[2]] # Case
            else:
               new_pos += "-"
            #print "Alt: ", pos, "Neu: ", new_pos
            res = new_pos

         elif pos == "ADJA":

            new_pos = "A-" # POS and type 
            new_pos += mapValDict['ADJA'][morphpos_splitted[1]]  #Degree
            if morphpos_splitted[4] != "*":
               new_pos += mapValDict['ADJA'][morphpos_splitted[4]] # Gender
            else:
               new_pos += "-"
            if morphpos_splitted[3] != "*":
               new_pos += mapValDict['ADJA'][morphpos_splitted[3]] # Number
            else:
               new_pos += "-"
            if morphpos_splitted[2] != "*":
               new_pos += mapValDict['ADJA'][morphpos_splitted[2]] # Case
            else:
               new_pos += "-"
            res = new_pos

         elif pos == "ART":
            new_pos = "T-" # POS
            new_pos += mapValDict['ART'][morphpos_splitted[1]] # Type
            if morphpos_splitted[4] != "*":
               new_pos += mapValDict['ART'][morphpos_splitted[4]] # Gender
            else:
               new_pos += "-"
            if morphpos_splitted[3] != "*":
               new_pos += mapValDict['ART'][morphpos_splitted[3]] # Number
            else:
               new_pos += "-"
            if morphpos_splitted[2] != "*":
               new_pos += mapValDict['ART'][morphpos_splitted[2]] # Case
            else:
               new_pos += "-"
            new_pos += "-" # Possesor (?)
            res = new_pos

         elif pos == "VFIN":
            new_pos = "V" # POS
            new_pos += mapValDict['VFIN'][morphpos_splitted[1]] # Type
            if len(morphpos_splitted) == 6:
               new_pos += mapValDict['VFIN'][morphpos_splitted[5]] # Mood
               new_pos += mapValDict['VFIN'][morphpos_splitted[4]] # Tense
               new_pos += mapValDict['VFIN'][morphpos_splitted[2]] # Person
               new_pos += mapValDict['VFIN'][morphpos_splitted[3]] # Number
               new_pos += "-" # Gender
            else:
               new_pos += "-----" # Mood, person, number, gender, tense
            res = new_pos

         elif pos == "VINF":
            new_pos = "V" # POS
            new_pos += mapValDict['VINF'][morphpos_splitted[1]] # Type
            if len(morphpos_splitted) == 3:
               new_pos += mapValDict['VINF'][morphpos_splitted[2]] # Mood "zu"
            else:
               new_pos += "n" # Mood "infinitive"
            new_pos += "-----" # Mood, person, number, gender, tense, clitic
            res = new_pos

         elif pos == "PRO":
            new_pos = "P" # POS
            if len(morphpos_splitted) == 6:
               new_pos += mapValDict['PRO'][morphpos_splitted[1]] # Type
               new_pos += "-" # Person
               new_pos += mapValDict['PRO'][morphpos_splitted[5]] # Gender
               new_pos += mapValDict['PRO'][morphpos_splitted[4]] # Number
               new_pos += mapValDict['PRO'][morphpos_splitted[3]] # Case
               new_pos += "-" # Possessor
            elif len(morphpos_splitted) == 7:
               new_pos += mapValDict['PRO'][morphpos_splitted[1]] # Type
               new_pos += mapValDict['PRO'][morphpos_splitted[3]] # Person
               new_pos += "-" # Gender
               new_pos += mapValDict['PRO'][morphpos_splitted[5]] # Number
               new_pos += mapValDict['PRO'][morphpos_splitted[4]] # Case
               new_pos += "-" # Possessor
            else:
               print "Pronomen nicht ok ", morphpos_splitted
               new_pos += "----" # Person, gender, number, case, possessor
               
            res = new_pos

         elif pos == "CONJ":
            new_pos = "C" # POS
            if len(morphpos_splitted) >= 2:
               new_pos += mapValDict['CONJ'][morphpos_splitted[1]] # Type
            else:
               print "Konjunktion nicht ok ", morphpos_splitted
               new_pos += "-" # Type
               
            res = new_pos

         elif pos == "PART":
            new_pos = "Q" # POS
            if len(morphpos_splitted) >= 2:
               new_pos += mapValDict['PART'][morphpos_splitted[1]] # Type
            else:
               print "Partikel nicht ok ", morphpos_splitted
               new_pos += "-" # Type
               
            res = new_pos
            

         else:
            res = mapValDict[pos]
            #print "Alt: ", pos, "Neu: ", new_pos

      else:

         #print "Hier"

         if pos == "VIMP":
            new_pos = "V" # POS
            new_pos += mapValDict['VFIN'][morphpos_splitted[1]] # Type
            new_pos += "m-" # Mood: imperative, tense
            new_pos += mapValDict['VFIN'][morphpos_splitted[2]] # Person
            new_pos += mapValDict['VFIN'][morphpos_splitted[3]] # Number
            new_pos += "-" # Gender
            res = new_pos

         elif pos == "VPP":
            new_pos = "V" # POS
            new_pos += mapValDict['VFIN'][morphpos_splitted[1]] # Type
            new_pos += "p----" # Mood, tense, person, number, gender, tense
            res = new_pos

         elif pos == "TRUNC":
            #print morphpos_splitted
            new_pos = ""
            if len(morphpos_splitted) > 1:
               #print "T: ", morphpos_splitted
               if morphpos_splitted[1] == "Noun":
                  new_pos = "N----" 
                  res = new_pos
               elif morphpos_splitted[1] == "Adj":
                  new_pos = "A----"
                  res = new_pos
               elif morphpos_splitted[1] == "Verb":
                  new_pos = "V------"
                  res = new_pos
            else:
               #print "TRUNC unklar ", morphpos_splitted
               res = "X"
            
         else:
            print "Unbekanntes POS: ", pos

      if len(res) > 1:
         if (res[0] == res[0].upper() and res[0] != "-") and (res[1] == res[1].upper() and res[1] != "-"):
            print "Keine POS-Aenderung fuer ", res

      return res

   return posToMulText(morphpos)

# TreeTagger (STTS tag set) to MulText
def retagGermanTT(morphpos):

   def posToMulText(pos):
      
      res = ""
      
      mapValDict = {}
      
      mapPosDict = {'ADJA': 'A-----', 'ADJD': 'R--', 'ADV': 'R--', 'APPR': 'Rps', 'APPRART': 'Rpc', 'APPO': 'Rts', 'APZR': 'Scs', 'ART': 'T----', 'CARD': 'Ac----', 'FM': 'X', 'ITJ': 'I', 'KOUI': 'Ci', 'KOUS': 'C-', 'KON': 'Cc', 'KOKOM': 'Cv', 'NN': 'N----', 'NE': 'Np---', 'PDS': 'Dd-----', 'PDAT': 'Pd-----', 'PIS': 'Pi-----', 'PIAT': 'Pi-----', 'PIDAT': 'Pi-----', 'PPER': 'Pp-----', 'PPOSS': 'Ds-----', 'PPOSAT': 'Ds-----', 'PRELS': 'Pr-----', 'PRELAT': 'Pr-----', 'PRF': 'Px-----', 'PWS': 'Pt-----', 'PWAT': 'Pt-----', 'PWAV': 'Ri-', 'PAV': 'Ri-', 'PTKZU': 'Qi', 'PTKNEG': 'R--', 'PTKVZ': 'Qv', 'PTKANT': 'R--', 'PTKA': 'Qs', 'TRUNC': 'Nc---', 'VVFIN': 'Vmi-----', 'VVIMP': 'Vmm-----', 'VVINF': 'Vmn-----', 'VVIZU': 'Vmu-----', 'VVPP': 'Vmp-----', 'VAFIN': 'Vai-----', 'VAIMP': 'Vam-----', 'VAINF': 'Van-----', 'VAPP': 'Vap-----', 'VMFIN': 'Voi-----', 'VMINF': 'Von-----', 'VMPP': 'Vop-----', 'XY': 'X', '\$,': 'Fi', '\$. ': 'Fe', '\$(': 'Fi'}

      pos_group = pos[0]
      #print "Pos = ", pos
      res = pos

      if pos in mapPosDict:
         res = mapPosDict[pos]

      else:
   
         if pos_group in mapValDict:
         
            try:
               res = mapValDict[pos_group][pos]
            except KeyError:
               res = pos
    

      return res

   return posToMulText(morphpos)


def retagSpanish(pos):

   def posToMulText(pos):
      
      res = ""
      
      mapValDict = {}
      mapValDict['N'] = {'NC': 'Nc---', 'NP': 'Np---'}
      mapValDict['V'] = {'VEadj': 'Vap-----', 'VHadj': 'Vap-----', 'VSadj': 'Vap-----', 'VMadj': 'Vop-----', 'VLadj': 'Vmp-----', 'VEfin': 'Vai-----', 'VHfin': 'Vai-----', 'VMfin': 'Voi-----', 'VSfin': 'Vai-----', 'VMfin': 'Voi-----', 'VLfin': 'Vmi-----', 'VEger': 'Vag-----', 'VHger': 'Vag-----', 'VSger': 'Vag-----',  'VMger': 'Vog-----', 'VLger': 'Vmg-----', 'VEinf': 'Van-----', 'VHinf': 'Van-----', 'VSinf': 'Van-----', 'VMinf': 'Von-----', 'VLinf': 'Vmn-----', 'VCLIger': 'Vmg----t', 'VCLIfin': 'Vmg----t', 'VCLIinf': 'Vmg----t',}
      mapValDict['A'] = {'ADJ': 'A-----'}
      mapPosDict = {'ART': 'T---', 'ALFS': 'X', 'ALFP': 'X', 'ACRNM': 'X', 'ADV': 'R--'}
      #mapPosDict = {'SYM': 'S', 'APPR': 'Sp-', 'APPRART': 'Sp-', 'ADJD': 'R--', 'ADV': 'R--', 'CARD': 'X', 'FM': 'X', 'PROADV': 'R--', 'ITJ': 'I', 'APPO': 'St-', 'APZR': 'St-'}

      pos_group = pos[0]
      #print "Pos = ", pos
      res = pos

      if pos in mapPosDict:
         res = mapPosDict[pos]

      else:
   
         if pos_group in mapValDict:
         
            try:
               res = mapValDict[pos_group][pos]
            except KeyError:
               res = pos


      return res

   return posToMulText(pos)


def retagFrench(pos):

   def posToMulText(pos):
      
      res = ""
      
      mapValDict = {}
      mapValDict['N'] = {'NOM': 'Nc---', 'NAM': 'Np---'}
      mapValDict['V'] = {'VER:cond': 'V-c-----', 'VER:futu': 'V-if----', 'VER:impe': 'V-m-----', 'VER:impf': 'V-ii----', 'VER:infi': 'V-n-----', 'VER:pper': 'V-ps----', 'VER:ppre': 'V-pp----', 'VER:pres': 'V-ip----', 'VER:simp': 'V-is----', 'VER:subi': 'V-si----', 'VER:subp': 'V-sp----'}
      mapValDict['A'] = {'ADJ': 'A-----'}
      mapPosDict = {'ADV': 'R--', 'ABR': 'X'}

      pos_group = pos[0]
      #print "Pos = ", pos
      res = pos

      if pos in mapPosDict:

         res = mapPosDict[pos]
   
      elif pos_group in mapValDict:
         
         try:
            res = mapValDict[pos_group][pos]
         except KeyError:
            res = pos

      return res

   return posToMulText(pos)

def retagEnglish(pos):

   def posToMulText(pos):
      
      res = ""
      
      mapValDict = {}
      mapValDict['N'] = {'NN': 'Ncs-', 'NNS': 'Ncp-', 'NP': 'Nps-', 'NPS': 'Npp-'}
      mapValDict['V'] = {'VB': 'Van---', 'VBD': 'Vaid--', 'VBG': 'Vags--', 'VBN': 'Vapd--', 'VBP': 'Vais--', 'VBZ': 'Vaiss3', 'VV': 'Vmn---', 'VVD': 'Vmid--', 'VVG': 'Vmgs--', 'VVN': 'Vmpd--', 'VVP': 'Vmis--', 'VVZ': 'Vmiss3'}
      mapPosDict = {'JJ': 'Ap-', 'JJR': 'Ac-', 'JJS': 'As-'}

      pos_group = pos[0]
      #print "Pos = ", pos
      res = pos
         
      if pos in mapPosDict:
         res = mapPosDict[pos]

      elif pos_group in mapValDict:
         
         try:
            res = mapValDict[pos_group][pos]
         except KeyError:
            res = pos

      return res

   return posToMulText(pos)


def modifyDeLinesRF(rft_input):

   targetEncoding = "utf-8"
   rft_file = open(rft_input, "r")
   out_file = open(out, "w")

   rft_line = rft_file.readline()
   line_nr = 1

   # <s> symbol at the file beginning
   out_file.write("<s>\n")
   last_line = rft_line

   while rft_line:

      #print rft_line
      print line_nr, "\r",
      
      try:
          if rft_line == "\n":
             out_file.write("</s>\n")
          else:

             if last_line == "\n":
                out_file.write("<s>\n")

             #print line_nr, ":", rft_line.strip().split("\t")
                
             rft_line = rft_line.strip()

             if not rft_line.startswith("<") and not rft_line.endswith("<"):

                (wf, pos, lemma) = rft_line.strip().split("\t")

               # Without unknowns
                if unknown == "no":
                   if lemma == "<unknown>":
                      lemma = wf

                new_line = wf + "\t" + retagGermanRF(pos) + "\t" + lemma

                out_file.write(new_line + "\n")

             else:
                #print "Meta kopieren"
                out_file.write(rft_line.strip() + "\n")
                

          last_line = rft_line

          rft_line = rft_file.readline()
          line_nr += 1
      
      except Exception as inst:
         print "Error at line ",  line_nr, " : ", rft_line, " = ", inst
         exit

   rft_file.close()
   out_file.close()

def modifyDeLinesTT(tt_input):

   targetEncoding = "utf-8"
   tt_file = open(tt_input, "r")
   out_file = open(out, "w")

   tt_line = tt_file.readline().strip()
   line_nr = 1

   while tt_line:

      #print rft_line
      print line_nr, "\r",
      
      try:

         # Metadata
          if tt_line.startswith("<") and tt_line.endswith(">"):
             out_file.write(tt_line + "\n")
          else:

             (wf, pos, lemma) = tt_line.strip().split("\t")

               # Without unknowns
             if unknown == "no":
                if lemma == "<unknown>":
                   lemma = wf

             new_line = wf + "\t" + retagGermanTT(pos) + "\t" + lemma

             out_file.write(new_line + "\n")

          tt_line = tt_file.readline().strip()
          line_nr += 1
      
      except Exception as inst:
         print "Fehler bei ",  line_nr, " : ", tt_line, " = ", inst
         exit

   tt_file.close()
   out_file.close()


# Input: Output of the Spanish TreeTagger:
# word \t POS \t [...] \t pos \s lemma
def modifyEsLines(tt_input):

   targetEncoding = "utf-8"
   tt_file = open(rft_input, "r")
   out_file = open(out, "w")

   tt_line = tt_file.readline().strip()
   line_nr = 1

   while tt_line:

      print line_nr, "\r",
      
      try:

         # Metadata
         if tt_line.startswith("<") and tt_line.endswith(">"):
            out_file.write(tt_line + "\n")
         else:

            (wf, pos_hyp, ignore, pos_lemma) = tt_line.strip().split("\t")

            pos = pos_lemma.strip().split()[0]
            lemma = pos_lemma.strip().split()[1]

            if unknown == "no":
               if lemma == "<unknown>":
                  lemma = wf

            new_line = wf + "\t" + retagSpanish(pos) + "\t" + lemma
            out_file.write(new_line + "\n")
         
         tt_line = tt_file.readline().strip()
         line_nr += 1
      
      except Exception as inst:
         print "Error at line ",  line_nr, " : ", tt_line, " = ", inst
         exit

   tt_file.close()
   out_file.close()


# Input: Output of the French TreeTagger:
# word \t POS \t lemma
def modifyFrLines(tt_input):

   targetEncoding = "utf-8"
   tt_file = open(tt_input, "r")
   out_file = open(out, "w")

   tt_line = tt_file.readline().strip()
   line_nr = 1

   while tt_line:

      print line_nr, "\r",
      
      try:

         # Metadata
         if tt_line.startswith("<") and tt_line.endswith(">"):
            out_file.write(tt_line + "\n")
      
         else:

            (wf, pos, lemma) = tt_line.strip().split("\t")

            if unknown == "no":
               if lemma == "<unknown>":
                  lemma = wf

            new_line = wf + "\t" + retagFrench(pos) + "\t" + lemma.strip()
            out_file.write(new_line + "\n")
         
         tt_line = tt_file.readline().strip()
         line_nr += 1
      
      except Exception as inst:
         print "Error at ",  line_nr, " : ", tt_line, " = ", inst
         exit

   tt_file.close()
   out_file.close()


# Input: Output of the English TreeTagger:
# word \t POS \t lemma
def modifyEnLines(tt_input):

   targetEncoding = "utf-8"
   tt_file = open(tt_input, "r")
   out_file = open(out, "w")

   tt_line = tt_file.readline().strip()
   line_nr = 1

   while tt_line:

      print line_nr, "\r",
      
      try:

         if tt_line.startswith("<") and tt_line.endswith(">"):
            out_file.write(tt_line + "\n")
            
         else:

            (wf, pos, lemma) = tt_line.strip().split("\t")

            if unknown == "no":
               if lemma == "<unknown>":
                  lemma = wf

            new_line = wf + "\t" + retagEnglish(pos) + "\t" + lemma.strip()
            out_file.write(new_line + "\n")
         
         tt_line = tt_file.readline().strip()
         line_nr += 1
      
      except Exception as inst:
         print "Error at line ",  line_nr, " : ", tt_line, " = ", inst
         exit

   tt_file.close()
   out_file.close()

################# MAIN #########################
if __name__ == "__main__":

   getOpts(sys.argv[1:])

   # German: RFTagger/TreeTagger
   if lang == "de":

      if tagger == "rf":
         modifyDeLinesRF(rft_input)
      elif tagger == "tt":
         modifyDeLinesTT(rft_input)
      else:
         print "Tagger not known"
             
   # Spanish: TreeTagger
   elif lang == "es":
      modifyEsLines(rft_input)

   # French: TreeTagger
   elif lang == "fr":
      modifyFrLines(rft_input)

   # English: TreeTagger
   elif lang == "en":
      modifyEnLines(rft_input)
   
